
"""
Compute CO2/NO2 mass along line or within area.
"""

from scipy.constants import N_A
from shapely.geometry import LineString, Polygon

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize

import doas
import amrs.vis

import smartcarb.map

from . import misc


def compute_mass(data, curve, field):
    """
    """
    distances = np.arange(0, np.nanmax(data.xp), 2e3)

    mass, mass_std = integrate_fields(data, field, curve, distances,
                                      tmax=100e3)

    valids = (distances > dmin) & np.isfinite(fluxes) & np.isfinite(fluxes_std)
    valids &= (distances <= dmax)

    mask = ~valids

    return np.ma.array(mass, mask=mask), np.ma.array(mass_std, mask=mask)



def integrate_fields(data, tracer, curve, distances, tmax):

    mass = []
    mass_std = []

    for d in distances:
        f,u = integrate_field(data, tracer, curve, d, tmax=tmax, show=False)
        mass.append(f)
        mass_std.append(u)

    return np.array(fluxes), np.array(fluxes_std)



def integrate_along_line(data, tracer, distances, curve=None, show=False,
                         method='fit'):
    """
    distance from source origin
    """

    masses = []
    masses_std = []
    fluxes = []

    wind_eff = []
    wind_10m = []
    areas = []

    if show:
        if tracer == 'CO2':
            fig = smartcarb.map.make_level2_map(data.lonc, data.latc,
                                                1e3 * data[tracer],
                                                label='CO$_2$ mass (g m$^{-2}$)',
                                                vmin=0, vmax=3)
        else:
            fig = smartcarb.map.make_level2_map(data.lonc, data.latc,
                                                1e6 * data[tracer],
                                                label='NO$_2$ mass (mg m$^{-2}$)',
                                                vmin=0, vmax=3)

        ax = fig.get_axes()[0]


    #fig = plt.figure()

    for i in range(distances.size-1):

        if method == 'fit':
            mass, mass_std, flux, pixels = fit_function(data, tracer, distances[i],
                                                        distances[i+1], -25e3, +25e3,
                                                        show=False)
        else:
            mass, mass_std, flux, pixels = extract_along_polygon(data, tracer,
                                                                 distances[i],
                                                                 distances[i+1],
                                                                 -25e3, 25e3, # FIXME
                                                                 curve=curve,
                                                                 show=show)
        masses.append(mass)
        masses_std.append(mass_std)
        fluxes.append(flux)
        areas.append(pixels)

        wind_eff.append(np.nanmean(
            np.sqrt(data['u_eff'].values[pixels]**2 + data['v_eff'].values[pixels]**2)
        ))
        wind_10m.append(np.nanmean(
            np.sqrt(data['u_10m'].values[pixels]**2 + data['v_10m'].values[pixels]**2)
        ))

        if show:

            t = np.linspace(-25e3, 25e3)
            x, y = curve.compute_normal(curve.t_o + distances[i], t=t)
            smartcarb.map.add_curve(ax, x, y, origin='Berlin', color='white')

            x, y = curve.compute_normal(curve.t_o + distances[i+1], t=t)
            smartcarb.map.add_curve(ax, x, y, origin='Berlin', color='white')

            color = ['yellow', 'red'][i%2]
            smartcarb.map.add_scatter(ax,
                                      data.lon.values[pixels],
                                      data.lat.values[pixels],
                                      color=color)

    return np.array(masses), np.array(masses_std), np.array(fluxes), \
           np.array(wind_eff), np.array(wind_10m), np.array(areas)



def add_area(ax, curve, distances, delta=10e3):

    # across-plume distance
    y = np.array([-25e3, 25e3])

    # compute arc length for parameters: arc(t)
    t = np.linspace(0.0, curve.t_o + distances.max() + delta, 5001)
    arc = misc.compute_arc_length(curve, curve.t_o, t)

    for i in range(distances.size):

        # interpolate distance from arc(t)
        xmin, xmax = np.interp(
            [distances[i] - delta / 2, distances[i] + delta / 2],
            arc, t
        )

        # compute normals
        (x0, x1), (y0, y1) = curve.compute_normal(xmin, t=y)
        (x3, x2), (y3, y2) = curve.compute_normal(xmax, t=y)

        # plot box
        xx,yy = np.array([x0, x1, x2, x3, x0]), np.array([y0, y1, y2, y3, y0])
        smartcarb.map.add_curve(ax, xx, yy, origin='Berlin', ls='-', color='yellow')



def fit_function(data, tracer, xa, xb, ya, yb, show=False, function='gauss'):

    if function == 'gauss':
        def func(x, *a):
            return a[0] + a[1] * x + a[2] * np.exp(-4.0 * np.log(2) * (x - a[3])**2 /
                                                   a[4]**2 )
    else:
        raise ValueError



    # mask
    x = np.array(data.xp)
    y = np.array(data.yp)
    pixels = (xa <= x) & (x < xb) & (ya <= y) & (y <= yb)

    # pixels
    y = y[pixels]
    c = data[tracer].values[pixels]

    # sort
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]

    # remove nans
    valids = np.isfinite(c)
    y = y[valids]
    c = c[valids]

    p0 = np.array([np.mean(c), 0.0, np.mean(c), 0.0, 10e3])

    try:
        p, cov_p = scipy.optimize.curve_fit(func, y, c, p0)
        integral = 0.5 * np.sqrt( np.pi / np.log(2) ) * p[2] * p[4]
        integral_std = np.nan # TODO
    except RuntimeError:
        raise
        p = None
        integral = np.nan
        integral_std = np.nan




    if show:
        noise = data[tracer].attrs.get('noise_level', 0.0)

        if noise == 'high':
            noise = 0.0

        plt.figure()
        if p is not None:
            plt.plot(y/1e3, func(y, *p), 'k-')
        plt.errorbar(y/1e3, c, noise, marker='.', ls='')


    return integral, integral_std, np.nan, pixels







def extract_along_polygon(data, tracer, xa, xb, ya, yb, curve=None, show=False):
    """
    Extract all pixels between on the interval [a,b[ using the pixel centers.
    """
    # mask
    x = np.array(data.xp)
    y = np.array(data.yp)
    pixels = (xa <= x) & (x < xb) & (ya <= y) & (y <= yb)

    # only use detected pixels (otherwise we might add up signals from other
    # plumes)
    #pixels &= data.detected_plume.values # FIXME

    # number of expected pixels in area
    n_expected = (xb - xa) * (yb - ya) / 2000**2

    n_pixels = np.sum(pixels)
    n_valids = np.sum(np.isfinite(data[tracer].values[pixels]))


    # TODO: check number of pixels in sub regions


    # pixels corners
    xc = np.array(data.xc)[pixels]
    yc = np.array(data.yc)[pixels]

    # trace gas mass
    c = np.array(data[tracer])[pixels]

    # wind
    if 'u_10m' in data:
        u10 = np.array(data['u_10m'])[pixels]
        v10 = np.array(data['v_10m'])[pixels]

    if 'u_eff' in data:
        u = np.array(data['u_eff'])[pixels]
        v = np.array(data['v_eff'])[pixels]

    # compute normal wind speed at pixel
    if curve is not None:
        t0 = curve.get_parameter(data.x.values[pixels], data.y.values[pixels])
        winds = []

        for i,(uu,vv) in enumerate(zip(u, v)):
            wind = np.dot([uu,vv], curve.compute_tangent(t0=t0[i], norm=True))
            winds.append(wind)

        winds = np.array(winds)


    if n_valids < 0.5 * n_expected:
        print(n_valids, n_pixels, n_expected)
        mass = np.nan
        mass_std = np.nan
        flux = np.nan

    else:
        # pixels
        polygons = [Polygon(np.transpose(x)) for x in zip(xc,yc)]

        # areas
        areas = np.array([px.area for px in polygons])

        # integral in kg/m
        mass = np.nansum(c * areas) / (xb - xa)

        # estimate noise
        noise_level = data[tracer].attrs.get('noise_level', np.nan)

        if isinstance(noise_level, str):
            molar_mass = 46.0055e-3 # of NO2
            mass_std = {
                'low':  1e15 * molar_mass / N_A * 1e4, # molec cm-2 -> kg m-2
                'high': 2e15 * molar_mass / N_A * 1e4,
                'S5':   1.3e15 * molar_mass / N_A * 1e4
            }[noise_level]
        else:
            mass_std = np.sum(noise_level * areas) / (xb - xa) / np.sqrt(np.sum(pixels))


        # flux using true wind speed and direction
        flux = np.sum(winds * c * areas) / (xb - xa)

    if False: #show:
        #plt.figure()
        plt.plot(y[pixels] / 1e3, c, '.', label='%d km' % ((xa+xb)/2e3))
        plt.xlabel('across-plume distance (km)')
        plt.ylabel('kg m-2')

        #plt.figure()
        #
        #plt.plot(y[pixels] / 1e3, np.sqrt(u**2 + v**2), 'o', label='eff. wind')
        #plt.plot(y[pixels] / 1e3, np.sqrt(u10**2 + v10**2), 'v', label='10m wind')
        #plt.plot(y[pixels] / 1e3, winds, 's', label='proj. wind')
        #plt.xlabel('across-plume distance (km)')
        #plt.legend()

    return mass, mass_std, flux, pixels




def extract_along_line(data, tracer, curve, t0, show=False, use_signal=False):
    """
    Extracts pixels along line. The line is computed as perpendicular to curve
    at distance t0.
    """
    if isinstance(tracer, str):
        tracer = data[tracer]

    values = np.array(tracer)[data.plume_area]

    try:
        error = tracer.attrs['noise_level']
    except (AttributeError, KeyError):
        error = np.nan

    # gaps in detected plume
    gaps = np.isnan(np.array(data.CO2)[data.plume_area])

    xcurve,ycurve = curve.compute_normal(t0=t0, t=[-50e3, 50e3])
    line = LineString(np.transpose([xcurve,ycurve]))

    # pixel centers
    x = np.array(data.x)[data.plume_area]
    y = np.array(data.y)[data.plume_area]

    # pixels corners
    xc = np.array(data.xc)[data.plume_area]
    yc = np.array(data.yc)[data.plume_area]

    # pixels
    pixels = [Polygon(np.transpose(x)) for x in zip(xc,yc)]

    pixels_x = []
    lengths_x = []
    values_x = []
    distances_x = []
    gaps_x = []

    for i,px in enumerate(pixels):

        if px.intersects(line) and np.isfinite(values[i]):

            isect = px.intersection(line)

            # distance of isect line from origin
            xi, yi = np.array(isect.coords.xy)[:,0]
            d = np.sqrt((xi-xcurve[0])**2 + (yi-ycurve[0])**2)

            pixels_x.append(px)
            values_x.append(values[i])
            distances_x.append(d)
            lengths_x.append(isect.length)
            gaps_x.append(gaps[i])


    if show:
        x0,y0 = 1e-3 * np.array(curve.compute_normal(t0=t0, t=[0.0]))

        fig = plt.figure(figsize=(8,3.5))
        ax = fig.add_subplot(121, aspect='equal')

        for i,px in enumerate(pixels):
            label = 'satellite pixel' if i == 0 else None
            xy = 1e-3 * np.array(px.exterior.coords.xy)
            ax.plot(*xy, '-', color='silver' if gaps[i] else 'gray',
                    label=label)

        for i,px in enumerate(pixels_x):
            label = 'pixel on line' if i == 0 else None
            xy = 1e-3 * np.array(px.exterior.coords.xy)
            ax.plot(*xy, '-', color='black', label=label)


        xy = 1e-3 * np.array(line.coords.xy)
        ax.plot(*xy, 'o-', color='#1f77b4', label='cross section')


        xc, yc = amrs.vis.corners2grid(data.xc, data.yc)
        tracer = np.array(tracer)
        #tracer[tracer < 0.05] = np.nan
        ax.pcolormesh(1e-3*xc,1e-3*yc, tracer, vmin=0, vmax=0.01, cmap='Reds')

        # center line
        xcc,ycc = 1e-3 * np.array(curve(t=curve.t_o+np.linspace(0, data.xp.max())))
        ax.plot(xcc,ycc,'-', color='#1f77b4')
        ax.plot(xcc[0],ycc[0],'D', color='#1f77b4')

        ax.grid(True)
        #ax.legend()

        r = max(np.abs(np.diff(xy[0])), np.abs(np.diff(xy[1])))

        ax.set_xlim(x0-0.55*r, x0+0.55*r)
        ax.set_ylim(y0-0.55*r, y0+0.55*r)

        ax.set_xlabel('x (km)')
        ax.set_ylabel('y (km)')
        ax.set_title('(a)')


        ax2 = fig.add_subplot(122)

        for d,l,v,g in zip(distances_x, lengths_x, values_x, gaps_x):
            t = 1e-3 * np.array([d,d+l]) - 50.0
            v = 1e3 * np.array([v,v])

            if not use_signal:
                ax2.fill_between(t, v-1e3*error, v+1e3*error, color='#1f77b4')

            if g:
                ax2.plot(t, v, '-', color='gray')
            else:
                ax2.plot(t, v, 'k-')


        ax2.set_xlabel('across plume (km)')
        ax2.set_ylabel('CO$_2$ mass (g$\,$m$^{-2}$)')

        xlim = np.max(np.abs(ax2.get_xlim()))
        ax2.set_xlim(-xlim, xlim)

        ax2.grid(True)
        ax2.set_title('(b)')

        plt.tight_layout()


    # sort values and lengths
    sorting = np.argsort(distances_x)

    values_x = np.array(values_x)[sorting]
    lengths_x = np.array(lengths_x)[sorting]
    gaps_x = np.array(gaps_x)[sorting]

    return values_x, lengths_x, gaps_x
